@@ -35,6 +35,14 @@ fn test_config_list_single() {
     insta::assert_snapshot!(stdout, @r###"
     test-table.somekey="some value"
     "###);
+
+    let stdout = test_env.jj_cmd_success(
+        test_env.env_root(),
+        &["config", "list", r#"-Tname ++ "\n""#, "test-table.somekey"],
+    );
+    insta::assert_snapshot!(stdout, @r###"
+    test-table.somekey
+    "###);
 }
 
 #[test]
@@ -114,6 +122,7 @@ fn test_config_list_all() {
     y.bar = 123
     "#,
     );
+
     let stdout = test_env.jj_cmd_success(test_env.env_root(), &["config", "list"]);
     insta::assert_snapshot!(
         find_stdout_lines(r"(test-val|test-table\b[^=]*)", &stdout),
@@ -245,6 +254,21 @@ fn test_config_layer_override_default() {
     # merge-tools.vimdiff.program="repo"
     merge-tools.vimdiff.program="command-arg"
     "###);
+
+    let stdout = test_env.jj_cmd_success(
+        &repo_path,
+        &[
+            "config",
+            "list",
+            "--color=always",
+            config_key,
+            "--include-overridden",
+        ],
+    );
+    insta::assert_snapshot!(stdout, @r###"
+    [38;5;8m# merge-tools.vimdiff.program="user"[39m
+    [38;5;2mmerge-tools.vimdiff.program[39m=[38;5;3m"repo"[39m
+    "###);
 }
 
 #[test]
