@@ -17,19 +17,19 @@
 use std::cmp::{max, min, Ordering};
 use std::collections::{BTreeSet, BinaryHeap, HashSet};
 use std::iter;
-use std::sync::Arc;
+use std::sync::{Arc, Mutex};
 
 use itertools::Itertools;
 
 use super::entry::{
     IndexEntry, IndexPosition, IndexPositionByGeneration, LocalPosition, SmallIndexPositionsVec,
+    SmallLocalPositionsVec,
 };
 use super::readonly::ReadonlyIndexSegment;
-use super::rev_walk::RevWalk;
+use super::rev_walk::{AncestorsBitSet, RevWalk};
 use super::revset_engine;
 use crate::backend::{ChangeId, CommitId};
 use crate::hex_util;
-use crate::id_prefix::{IdIndex, IdIndexSource, IdIndexSourceEntry};
 use crate::index::{AllHeadsForGcUnsupported, ChangeIdIndex, Index};
 use crate::object_id::{HexPrefix, ObjectId, PrefixResolution};
 use crate::revset::{ResolvedExpression, Revset, RevsetEvaluationError};
@@ -44,7 +44,7 @@ pub(super) trait IndexSegment: Send + Sync {
 
     fn name(&self) -> Option<String>;
 
-    fn commit_id_to_pos(&self, commit_id: &CommitId) -> Option<IndexPosition>;
+    fn commit_id_to_pos(&self, commit_id: &CommitId) -> Option<LocalPosition>;
 
     /// Suppose the given `commit_id` exists, returns the previous and next
     /// commit ids in lexicographical order.
@@ -55,6 +55,16 @@ pub(super) trait IndexSegment: Send + Sync {
 
     fn resolve_commit_id_prefix(&self, prefix: &HexPrefix) -> PrefixResolution<CommitId>;
 
+    fn resolve_neighbor_change_ids(
+        &self,
+        change_id: &ChangeId,
+    ) -> (Option<ChangeId>, Option<ChangeId>);
+
+    fn resolve_change_id_prefix(
+        &self,
+        prefix: &HexPrefix,
+    ) -> PrefixResolution<(ChangeId, SmallLocalPositionsVec)>;
+
     fn generation_number(&self, local_pos: LocalPosition) -> u32;
 
     fn commit_id(&self, local_pos: LocalPosition) -> CommitId;
@@ -157,9 +167,19 @@ impl<'a> CompositeIndex<'a> {
             .unwrap()
     }
 
+    pub fn entry_by_id(&self, commit_id: &CommitId) -> Option<IndexEntry<'a>> {
+        self.ancestor_index_segments().find_map(|segment| {
+            let local_pos = segment.commit_id_to_pos(commit_id)?;
+            let pos = IndexPosition(local_pos.0 + segment.num_parent_commits());
+            Some(IndexEntry::new(segment, pos, local_pos))
+        })
+    }
+
     pub fn commit_id_to_pos(&self, commit_id: &CommitId) -> Option<IndexPosition> {
-        self.ancestor_index_segments()
-            .find_map(|segment| segment.commit_id_to_pos(commit_id))
+        self.ancestor_index_segments().find_map(|segment| {
+            let LocalPosition(local_pos) = segment.commit_id_to_pos(commit_id)?;
+            Some(IndexPosition(local_pos + segment.num_parent_commits()))
+        })
     }
 
     /// Suppose the given `commit_id` exists, returns the previous and next
@@ -179,9 +199,68 @@ impl<'a> CompositeIndex<'a> {
             .unwrap()
     }
 
-    pub fn entry_by_id(&self, commit_id: &CommitId) -> Option<IndexEntry<'a>> {
-        self.commit_id_to_pos(commit_id)
-            .map(|pos| self.entry_by_pos(pos))
+    /// Suppose the given `change_id` exists, returns the minimum prefix length
+    /// to disambiguate it within all the indexed ids including hidden ones.
+    pub(super) fn shortest_unique_change_id_prefix_len(&self, change_id: &ChangeId) -> usize {
+        let (prev_id, next_id) = self.resolve_neighbor_change_ids(change_id);
+        itertools::chain(prev_id, next_id)
+            .map(|id| hex_util::common_hex_len(change_id.as_bytes(), id.as_bytes()) + 1)
+            .max()
+            .unwrap_or(0)
+    }
+
+    /// Suppose the given `change_id` exists, returns the previous and next
+    /// change ids in lexicographical order. The returned change ids may be
+    /// hidden.
+    pub(super) fn resolve_neighbor_change_ids(
+        &self,
+        change_id: &ChangeId,
+    ) -> (Option<ChangeId>, Option<ChangeId>) {
+        self.ancestor_index_segments()
+            .map(|segment| segment.resolve_neighbor_change_ids(change_id))
+            .reduce(|(acc_prev_id, acc_next_id), (prev_id, next_id)| {
+                (
+                    acc_prev_id.into_iter().chain(prev_id).max(),
+                    acc_next_id.into_iter().chain(next_id).min(),
+                )
+            })
+            .unwrap()
+    }
+
+    /// Resolves the given change id `prefix` to the associated entries. The
+    /// returned entries may be hidden.
+    ///
+    /// The returned index positions are sorted in ascending order.
+    pub(super) fn resolve_change_id_prefix(
+        &self,
+        prefix: &HexPrefix,
+    ) -> PrefixResolution<(ChangeId, SmallIndexPositionsVec)> {
+        use PrefixResolution::*;
+        self.ancestor_index_segments()
+            .fold(NoMatch, |acc_match, segment| {
+                if acc_match == AmbiguousMatch {
+                    return acc_match; // avoid checking the parent file(s)
+                }
+                let to_global_pos = {
+                    let num_parent_commits = segment.num_parent_commits();
+                    move |LocalPosition(pos)| IndexPosition(pos + num_parent_commits)
+                };
+                // Similar to PrefixResolution::plus(), but merges matches of the same id.
+                match (acc_match, segment.resolve_change_id_prefix(prefix)) {
+                    (NoMatch, local_match) => local_match.map(|(id, positions)| {
+                        (id, positions.into_iter().map(to_global_pos).collect())
+                    }),
+                    (acc_match, NoMatch) => acc_match,
+                    (AmbiguousMatch, _) => AmbiguousMatch,
+                    (_, AmbiguousMatch) => AmbiguousMatch,
+                    (SingleMatch((id1, _)), SingleMatch((id2, _))) if id1 != id2 => AmbiguousMatch,
+                    (SingleMatch((id, mut acc_positions)), SingleMatch((_, local_positions))) => {
+                        acc_positions
+                            .insert_many(0, local_positions.into_iter().map(to_global_pos));
+                        SingleMatch((id, acc_positions))
+                    }
+                }
+            })
     }
 
     pub(super) fn is_ancestor_pos(
@@ -417,50 +496,62 @@ impl Index for CompositeIndex<'_> {
 
 pub(super) struct ChangeIdIndexImpl<I> {
     index: I,
-    pos_by_change: IdIndex<ChangeId, IndexPosition, 4>,
+    reachable_set: Mutex<AncestorsBitSet>,
 }
 
 impl<I: AsCompositeIndex> ChangeIdIndexImpl<I> {
     pub fn new(index: I, heads: &mut dyn Iterator<Item = &CommitId>) -> ChangeIdIndexImpl<I> {
-        let mut pos_by_change = IdIndex::builder();
         let composite = index.as_composite();
-        let head_positions = heads
-            .map(|id| composite.commit_id_to_pos(id).unwrap())
-            .collect_vec();
-        for entry in composite.walk_revs(&head_positions, &[]) {
-            pos_by_change.insert(&entry.change_id(), entry.position());
+        let mut reachable_set = AncestorsBitSet::with_capacity(composite.num_commits());
+        for id in heads {
+            reachable_set.add_head(composite.commit_id_to_pos(id).unwrap());
         }
-        Self {
+        ChangeIdIndexImpl {
             index,
-            pos_by_change: pos_by_change.build(),
+            reachable_set: Mutex::new(reachable_set),
         }
     }
 }
 
 impl<I: AsCompositeIndex + Send + Sync> ChangeIdIndex for ChangeIdIndexImpl<I> {
+    /// Resolves change id prefix among all ids, then filters out hidden
+    /// entries.
+    ///
+    /// If `SingleMatch` is returned, the commits including in the set are all
+    /// visible. `AmbiguousMatch` may be returned even if the prefix is unique
+    /// within the visible entries.
     fn resolve_prefix(&self, prefix: &HexPrefix) -> PrefixResolution<Vec<CommitId>> {
-        self.pos_by_change
-            .resolve_prefix_with(self.index.as_composite(), prefix, |entry| entry.commit_id())
-            .map(|(_, commit_ids)| commit_ids)
+        let index = self.index.as_composite();
+        match index.resolve_change_id_prefix(prefix) {
+            PrefixResolution::NoMatch => PrefixResolution::NoMatch,
+            PrefixResolution::SingleMatch((_change_id, positions)) => {
+                debug_assert!(positions.iter().tuple_windows().all(|(a, b)| a < b));
+                let mut reachable_set = self.reachable_set.lock().unwrap();
+                reachable_set.visit_until(index, *positions.first().unwrap());
+                let reachable_commit_ids = positions
+                    .iter()
+                    .filter(|&&pos| reachable_set.contains(pos))
+                    .map(|&pos| index.entry_by_pos(pos).commit_id())
+                    .collect_vec();
+                if reachable_commit_ids.is_empty() {
+                    PrefixResolution::NoMatch
+                } else {
+                    PrefixResolution::SingleMatch(reachable_commit_ids)
+                }
+            }
+            PrefixResolution::AmbiguousMatch => PrefixResolution::AmbiguousMatch,
+        }
     }
 
+    /// Calculates the shortest prefix length of the given `change_id` among all
+    /// ids including hidden entries.
+    ///
+    /// The returned length is usually a few digits longer than the minimum
+    /// length to disambiguate within the visible entries.
     fn shortest_unique_prefix_len(&self, change_id: &ChangeId) -> usize {
-        self.pos_by_change
-            .shortest_unique_prefix_len(self.index.as_composite(), change_id)
-    }
-}
-
-impl<'index> IdIndexSource<IndexPosition> for CompositeIndex<'index> {
-    type Entry = IndexEntry<'index>;
-
-    fn entry_at(&self, pointer: &IndexPosition) -> Self::Entry {
-        self.entry_by_pos(*pointer)
-    }
-}
-
-impl IdIndexSourceEntry<ChangeId> for IndexEntry<'_> {
-    fn to_key(&self) -> ChangeId {
-        self.change_id()
+        self.index
+            .as_composite()
+            .shortest_unique_change_id_prefix_len(change_id)
     }
 }
 
