@@ -167,26 +167,13 @@ impl ReadonlyRepo {
         fs::write(&op_store_type_path, op_store.name()).context(&op_store_type_path)?;
         let op_store: Arc<dyn OpStore> = Arc::from(op_store);
 
-        let mut root_view = op_store::View::default();
-        root_view.head_ids.insert(store.root_commit_id().clone());
-
         let op_heads_path = repo_path.join("op_heads");
         fs::create_dir(&op_heads_path).context(&op_heads_path)?;
-        let operation_metadata =
-            crate::transaction::create_op_metadata(user_settings, "initialize repo".to_string());
-        let root_view_id = op_store.write_view(&root_view).unwrap();
-        let init_operation = op_store::Operation {
-            view_id: root_view_id,
-            parents: vec![op_store.root_operation_id().clone()],
-            metadata: operation_metadata,
-        };
-        let init_operation_id = op_store.write_operation(&init_operation).unwrap();
-        let init_operation = Operation::new(op_store.clone(), init_operation_id, init_operation);
         let op_heads_store = op_heads_store_initializer(user_settings, &op_heads_path);
-        op_heads_store.update_op_heads(&[], init_operation.id());
         let op_heads_type_path = op_heads_path.join("type");
         fs::write(&op_heads_type_path, op_heads_store.name()).context(&op_heads_type_path)?;
-        let op_heads_store = Arc::from(op_heads_store);
+        op_heads_store.update_op_heads(&[], op_store.root_operation_id());
+        let op_heads_store: Arc<dyn OpHeadsStore> = Arc::from(op_heads_store);
 
         let index_path = repo_path.join("index");
         fs::create_dir(&index_path).context(&index_path)?;
@@ -203,20 +190,33 @@ impl ReadonlyRepo {
             .context(&submodule_store_type_path)?;
         let submodule_store = Arc::from(submodule_store);
 
-        let view = View::new(root_view);
-        Ok(Arc::new(ReadonlyRepo {
+        let root_operation_data = op_store
+            .read_operation(op_store.root_operation_id())
+            .expect("failed to read root operation");
+        let root_operation = Operation::new(
+            op_store.clone(),
+            op_store.root_operation_id().clone(),
+            root_operation_data,
+        );
+        let root_view = root_operation.view().expect("failed to read root view");
+        let repo = Arc::new(ReadonlyRepo {
             repo_path,
             store,
             op_store,
             op_heads_store,
-            operation: init_operation,
+            operation: root_operation,
             settings: repo_settings,
             index_store,
             index: OnceCell::new(),
             change_id_index: OnceCell::new(),
-            view,
+            view: root_view,
             submodule_store,
-        }))
+        });
+        let mut tx = repo.start_transaction(user_settings);
+        tx.mut_repo()
+            .add_head(&repo.store().root_commit())
+            .expect("failed to add root commit as head");
+        Ok(tx.commit("initialize repo"))
     }
 
     pub fn loader(&self) -> RepoLoader {
@@ -771,11 +771,12 @@ impl MutableRepo {
             && self.view() == &self.base_repo.view)
     }
 
-    pub fn consume(self) -> (Box<dyn MutableIndex>, View) {
+    pub(crate) fn consume(self) -> (Box<dyn MutableIndex>, View) {
         self.view.ensure_clean(|v| self.enforce_view_invariants(v));
         (self.index, self.view.into_inner())
     }
 
+    /// Returns a [`CommitBuilder`] to write new commit to the repo.
     pub fn new_commit(
         &mut self,
         settings: &UserSettings,
@@ -785,6 +786,7 @@ impl MutableRepo {
         CommitBuilder::for_new_commit(self, settings, parents, tree_id)
     }
 
+    /// Returns a [`CommitBuilder`] to rewrite an existing commit in the repo.
     pub fn rewrite_commit(
         &mut self,
         settings: &UserSettings,
@@ -795,7 +797,8 @@ impl MutableRepo {
         // `self.rewritten_commits`
     }
 
-    pub fn write_commit(
+    /// Only called from [`CommitBuilder::write`]. Use that function instead.
+    pub(crate) fn write_commit(
         &mut self,
         commit: backend::Commit,
         sign_with: Option<&mut SigningFn>,
@@ -1370,15 +1373,15 @@ impl Repo for MutableRepo {
         self.index.as_index()
     }
 
-    fn submodule_store(&self) -> &Arc<dyn SubmoduleStore> {
-        self.base_repo.submodule_store()
-    }
-
     fn view(&self) -> &View {
         self.view
             .get_or_ensure_clean(|v| self.enforce_view_invariants(v))
     }
 
+    fn submodule_store(&self) -> &Arc<dyn SubmoduleStore> {
+        self.base_repo.submodule_store()
+    }
+
     fn resolve_change_id_prefix(&self, prefix: &HexPrefix) -> PrefixResolution<Vec<CommitId>> {
         let change_id_index = self.index.change_id_index(&mut self.view().heads().iter());
         change_id_index.resolve_prefix(prefix)
